// C Source File
// Created 8/5/00; 5:53:26 AM

// Available Patches
// #define USE_GRAY_PATCH
// #define USE_LONGMUL_PATCH
// #define USE_LONGDIV_PATCH
// #define USE_INITS_PATCH

#define SAVE_SCREEN  // Save/Restore LCD Contents

#include <nostub.h>  // Make NoStub Program
#include <all.h>     // Include All Header Files

#define Screen_Height 99
#define Screen_Width 159
#define Start_X 10
#define Start_Y 10

int _ti89;           // Produce .89Z File
int _ti92plus;       // Produce .9XZ File

// forward declarations
void title_screen();
void disable_interrupt();
void enable_interrupt();
void put_new_nibble(int x, int y);
void erase_nibble(int x, int y);
void initialize_screen();
void delay(unsigned int delay2);
void initialize_arrays();
void read_keypresses(char *dir);
void update_direction(char dir, int *x, int *y, BOOL *dead);
void test_for_nibble_on_tail(int length, int x, int y, BOOL *on_tail);
void gameplay();

// global variables
int player_x=8;
int player_y=2;
int nibble_x=0;
int nibble_y=0;
unsigned int delay_length=0;
unsigned int high_score=0;
unsigned char tail_x[850] = {};
unsigned char tail_y[850] = {};
char esc='n';

// external variables
extern void _old_interrupt1;
extern void _temporary_interrupt1;

void title_screen() {
	ClrScr();							
	DrawLine(0, 0, 0, Screen_Height,A_NORMAL);
	DrawLine(0, 0, Screen_Width, 0,A_NORMAL);
	DrawLine(Screen_Width, 0, Screen_Width, Screen_Height,A_NORMAL);
	DrawLine(0, Screen_Height, Screen_Width, Screen_Height,A_NORMAL);
	FontSetSys(F_8x10);				// set font to large size
	DrawStr(20,3,"C-Nibbles V1.0",A_NORMAL);
	DrawLine(0, 15, Screen_Width, 15,A_NORMAL);
	FontSetSys(F_6x8);				// set font to medium size
	DrawStr(10,20,"Speed:",A_NORMAL);
	DrawStr(10,68,"Other:",A_NORMAL);
	FontSetSys(F_4x6);				// set font to small size
	DrawStr(20,30,"F1. Very slow.",A_NORMAL);
	DrawStr(20,37,"F2. Slow.",A_NORMAL);
	DrawStr(20,44,"F3. Medium.",A_NORMAL);
	DrawStr(20,51,"F4. Fast.",A_NORMAL);
	DrawStr(20,58,"F5. Very fast.",A_NORMAL);
	DrawStr(20,78,"Alpha. About.",A_NORMAL);
	DrawStr(20,85,"Shift. Contact.",A_NORMAL);
	DrawStr(50,92,"Press ESC to Exit",A_NORMAL);
	delay_length=35000;				// the default delay should be medium speed
	do {
		if(_rowread(0xffbf) & 1) {	// check for escape to be pressed
			esc='y';				// skip the gameplay function
			break;				// break out of the do while loop			
		}
		if(_rowread(0xfffd) & 128) {	// check for f5
			delay_length=15000;
			break;	
		}
		if(_rowread(0xfffb) & 128) {	// check for f4
			delay_length=25000;
			break;
		}
		if(_rowread(0xfff7) & 128) {	// check for f3
			delay_length=35000;
			break;
		}
		if(_rowread(0xffef) & 128) {	// check for f2
			delay_length=45000;
			break;
		}
		if(_rowread(0xffdf) & 128) {	// check for f1
			delay_length=55000;
			break;
		}
		if(_rowread(0xfffe) & 128) {	// check for alpha
			do {
				DrawStr(80,30,"C-Nibbles is a simple C",A_NORMAL);
				DrawStr(80,36,"clone of the classic",A_NORMAL);
				DrawStr(80,42,"nibbles game.  I wrote",A_NORMAL);
				DrawStr(80,48,"this in order to learn",A_NORMAL);
				DrawStr(80,54,"C.  Thus, the value of",A_NORMAL);
				DrawStr(80,60,"C-Nibbles is the highly",A_NORMAL);
				DrawStr(80,66,"explained source, not",A_NORMAL);
				DrawStr(80,72,"the game itself.",A_NORMAL);
				DrawStr(85,80,"*Press Enter*",A_NORMAL);
			} while(!(_rowread(0xffd) & 1));
			title_screen();		// recursively call the function title_screen()
		}
		if(_rowread(0xfffe) & 32) {	// check for shift
			do {
				DrawStr(80,30,"My name is Matthew",A_NORMAL);
				DrawStr(80,36,"Landry and you can",A_NORMAL);
				DrawStr(80,42,"e-mail me at:",A_NORMAL);
				DrawStr(80,48,"mhlandry1@home.com",A_NORMAL);
				DrawStr(80,54,"I appreciate all",A_NORMAL);
				DrawStr(80,60,"insightful comments.",A_NORMAL);
				DrawStr(85,68,"*Press Enter*",A_NORMAL);
			} while(!(_rowread(0xffd) & 1));
			title_screen();
		}			
	} while(!(_rowread(0xfffe) & 16));	// repeat until 2nd is pressed
}

void disable_interrupt1() {			// replaces TIOS int1 with rte
	asm volatile("move.w #0x0700,%d0
     	trap #1
     	move.l (0x040064),(_old_interrupt1)
     	move.l #_temporary_interrupt1,(0x040064)
     	trap #1");
}

void enable_interrupt1() {			// restores TIOS's int1
	asm volatile("move.l #0x0700,%d0
     	trap #1
     	move.l (_old_interrupt1),(0x040064)
     	trap #1");
}

void put_new_nibble(int x, int y) {			
	int tempx, tempy;
	for(tempy=y; tempy<y+4; tempy++) {
		for(tempx=x-4; tempx<x; tempx++)
			DrawPix(tempx, tempy,A_NORMAL);	// draws a 4x4 nibble
	}
}

void erase_nibble(int x, int y) {
	int tempx, tempy;
	for(tempy=y; tempy<y+4; tempy++) {
		for(tempx=x-4; tempx<x; tempx++)
			DrawPix(tempx, tempy,A_REVERSE);	// erases a 4x4 nibble
	}
}

void initialize_screen() {
	int tempy;
	DrawLine(0, 0, 0, Screen_Height,A_THICK1);
	DrawLine(0, 0, Screen_Width, 0,A_NORMAL);
	DrawLine(Screen_Width-1, 0, Screen_Width-1, Screen_Height,A_THICK1);
	DrawLine(0, Screen_Height, Screen_Width, Screen_Height,A_NORMAL);
	for(tempy=player_y; tempy<player_y+16; tempy+=4) // draw initial tail
		put_new_nibble(player_x,tempy);
}

void delay(unsigned int delay2) {		// create a delay because the program runs too fast
	int temp, temp2;
	for(temp=0; temp<=delay2; temp++)
		temp2++;
}

void initialize_arrays() { 	// I do this manually because I am lazy
	tail_x[0]=(8);
	tail_x[1]=(8);
	tail_x[2]=(8);
	tail_x[3]=(8);
	tail_y[0]=(14);
	tail_y[1]=(10);
	tail_y[2]=(6);
	tail_y[3]=(2);
}

void read_keypresses(char *dir) {		// reads te keypresses and sets the direction accordingly
	char kbdata;
	switch(kbdata=_rowread(0x0ffee)) {
		case 1:
			(*dir)='u';
			break;
		case 2:
			(*dir)='l';
			break;
		case 4:
			(*dir)='d';
			break;
		case 8:
			(*dir)='r';
			break;
	}
}

void update_direction(char dir, int *x, int *y, BOOL *dead) {
// uses the variable direction to update the nibble
	switch(dir) {
		case 'r':
			if((*x)>=154)
				(*dead)=TRUE;
			(*x)+=4;
			break;
		case 'l':
			if((*x)<=8)
				(*dead)=TRUE;
			(*x)-=4;
			break;
		case 'u':
			if((*y)<=4)
				(*dead)=TRUE;
			(*y)-=4;
			break;
		case 'd':
			if((*y)>=92)
				(*dead)=TRUE;
			(*y)+=4;
			break;
	}
}

void test_for_nibble_on_tail(int length, int x, int y, BOOL *on_tail) {
// check to see if the nibble is placed on the tail
	int temp;
	(*on_tail)=FALSE;
	for(temp=0; temp<=length-1; temp++) {
		if(tail_x[temp]==x && tail_y[temp]==y) {
			(*on_tail)=TRUE;
			break;
			}
	}
}

void gameplay() {			// lets play!
	int temp;
	static unsigned int length=0, score=0;		// always, always initialise static variables
	static BOOL dead=FALSE;
	static char direction='r', nibble_status='n', skip_update='n';
     BOOL nibble_on_tail;
	dead=FALSE;
	nibble_on_tail=FALSE;
	length=4;
	player_x=8;
	player_y=2;
	score=0;
	direction='r';
	nibble_status='n';
	skip_update='n';
	randomize();
	initialize_arrays();
	initialize_screen();
	while (dead==FALSE) { 				// repeat until return 0 occurs
		if (_rowread(0xffbf) & 1)		// read from the row with escape, and test for escape
			dead=TRUE;				     // exit if escape is pressed	
		read_keypresses(&direction);
		update_direction(direction, &player_x, &player_y, &dead);
		// check status of nibble, place if necessary
		if(nibble_status=='n') {			// checks to see if a nibble is on the screen
			do {
				nibble_status='y';		// set the var to show that a nibble has been placed
				nibble_x=(random(38)*4)+8;	// get a random value for x
				nibble_y=(random(23)*4)+2;	// same for y
				test_for_nibble_on_tail(length, nibble_x, nibble_y, &nibble_on_tail);
				// check to see if the nibble is placed on the tail
			} while(nibble_on_tail==TRUE);
		}
		// put the nibble
		put_new_nibble(nibble_x, nibble_y);
		// erase last nibble from tail, make it look animated
		erase_nibble(tail_x[0],tail_y[0]);
		// adjust coordinates for animation
		if(skip_update=='n') {
			for(temp=0; temp<length-1; temp++) {
				tail_x[temp]=tail_x[temp+1];
				tail_y[temp]=tail_y[temp+1];
			}
		}
		else 
			skip_update='n';
		// put coordinates for new nibble, animation
		tail_x[length-1]=player_x;
		tail_y[length-1]=player_y;
		// check for eating a nibble
		if(tail_x[length-1]==nibble_x && tail_y[length-1]==nibble_y) {
			length++;			// length increases
			score++;			// score increases
			skip_update='y';	// skip an update, b/c we dont need to erase the last nibble on the tail
			erase_nibble(nibble_x, nibble_y);	// erase the nibble - 'food'
			nibble_status='n';		// we need to place another nibble
		}
		// put new nibble
		put_new_nibble(player_x,player_y);
		// check for death!
		for(temp=0; temp<length-2; temp++) {  // check to see if the tail has hit itself 
			if(tail_x[temp]==player_x && tail_y[temp]==player_y)
				dead=TRUE;
		}
		// delay
		delay(delay_length);
	}
	// this is the sequence that happens after you die or press esc
	ClrScr();
	DrawLine(0, 0, 0, Screen_Height,A_NORMAL);
	DrawLine(0, 0, Screen_Width, 0,A_NORMAL);
	DrawLine(Screen_Width, 0, Screen_Width, Screen_Height,A_NORMAL);
	DrawLine(0, Screen_Height, Screen_Width, Screen_Height,A_NORMAL);
	FontSetSys(F_8x10);
	DrawStr(20,3,"C-Nibbles V1.0",A_NORMAL);
	DrawLine(0, 15, Screen_Width, 15,A_NORMAL);
	FontSetSys(F_6x8);
	DrawStr(5,30,"Your score was:",A_NORMAL);
	if(score>high_score)
		DrawStr(20,50,"You got a high score!",A_NORMAL);
	temp=120;			// x value
	do {				// we need a good way to display a variable of integer type
		DrawChar(temp,30,score%10+'0',A_NORMAL);	// modular divide by ten, thus the remainder is 
											// a character bt. 0-9
		score/=10;							// actually divide by ten
		temp-=6;								// decrease the x coordinate for the next digit
	} while(score>0);							// continue until score = 0
	DrawStr(25,85,"Press ESC to exit",A_NORMAL);
	do {
		temp=0;
	} while(!(_rowread(0xffbf) & 1));		// repeat until esc is pressed
}



// Main Function
int _main() {
	esc='n';				// initialize the variable
	disable_interrupt1();	// install custom interrupt
     title_screen();		// call the title screen
     ClrScr();				// clear the screen
     if(esc=='n')			// if esc was hit during the title screen, dont run the gameplay 
		gameplay();
	enable_interrupt1();	// restore TIOS's interrupt
	return 0;					// exit back to tios
}

// Much, Much thanks to *Olle* and Patrick Davidson and SMN and Dan Englender
/* Much thanks goes to the following people for putting up with my constant questions:
Olle Hedman: 
	* Finding a major bug that caused the program to crash the second time it was run.
	* Showing me a very cool optimzation tip
	* Other smaller bug fixes
	
Dan Englender:
	* Kindly answering all of my constant, newbie questions and giving me helpful information about C
	  in general.
	  
SMN:
	* Sometimes kindly answering my questions.
	* Searching through my crappy code for bugs.
	
Patrick Davidson:
	* Kindly answering my questions.
	* Sharing some example source code with me.
	
*/